@@ -1,8 +1,8 @@
+import concurrent.futures
 import logging
 import os
 import re
 import sys
-import threading
 from getpass import getpass
 from hashlib import md5
 from string import Formatter
@@ -12,6 +12,7 @@ import click
 import requests
 from tqdm import tqdm
 
+from .bases import Track, Video
 from .clients import DeezerClient, QobuzClient, SoundCloudClient, TidalClient
 from .config import Config
 from .constants import (
@@ -19,18 +20,19 @@ from .constants import (
     DB_PATH,
     LASTFM_URL_REGEX,
     MEDIA_TYPES,
+    QOBUZ_INTERPRETER_URL_REGEX,
     SOUNDCLOUD_URL_REGEX,
     URL_REGEX,
 )
 from .db import MusicDB
-from .downloader import Album, Artist, Label, Playlist, Track, Tracklist
 from .exceptions import (
     AuthenticationError,
     NonStreamable,
     NoResultsFound,
     ParsingError,
 )
-from .utils import capitalize
+from .tracklists import Album, Artist, Label, Playlist, Tracklist
+from .utils import extract_interpreter_url
 
 logger = logging.getLogger(__name__)
 
@@ -41,6 +43,7 @@ MEDIA_CLASS = {
     "artist": Artist,
     "track": Track,
     "label": Label,
+    "video": Video,
 }
 Media = Union[Album, Playlist, Artist, Track]
 
@@ -54,6 +57,7 @@ class MusicDL(list):
         self.url_parse = re.compile(URL_REGEX)
         self.soundcloud_url_parse = re.compile(SOUNDCLOUD_URL_REGEX)
         self.lastfm_url_parse = re.compile(LASTFM_URL_REGEX)
+        self.interpreter_url_parse = re.compile(QOBUZ_INTERPRETER_URL_REGEX)
 
         self.config = config
         if self.config is None:
@@ -76,48 +80,6 @@ class MusicDL(list):
         else:
             self.db = []
 
-    def prompt_creds(self, source: str):
-        """Prompt the user for credentials.
-
-        :param source:
-        :type source: str
-        """
-        if source == "qobuz":
-            click.secho(f"Enter {capitalize(source)} email:", fg="green")
-            self.config.file[source]["email"] = input()
-            click.secho(
-                f"Enter {capitalize(source)} password (will not show on screen):",
-                fg="green",
-            )
-            self.config.file[source]["password"] = md5(
-                getpass(prompt="").encode("utf-8")
-            ).hexdigest()
-
-            self.config.save()
-            click.secho(f'Credentials saved to config file at "{self.config._path}"')
-        else:
-            raise Exception
-
-    def assert_creds(self, source: str):
-        assert source in (
-            "qobuz",
-            "tidal",
-            "deezer",
-            "soundcloud",
-        ), f"Invalid source {source}"
-        if source == "deezer":
-            # no login for deezer
-            return
-
-        if source == "soundcloud":
-            return
-
-        if source == "qobuz" and (
-            self.config.file[source]["email"] is None
-            or self.config.file[source]["password"] is None
-        ):
-            self.prompt_creds(source)
-
     def handle_urls(self, url: str):
         """Download a url
 
@@ -153,27 +115,40 @@ class MusicDL(list):
         item = MEDIA_CLASS[media_type](client=client, id=item_id)
         self.append(item)
 
+    def _get_download_args(self) -> dict:
+        return {
+            "database": self.db,
+            "parent_folder": self.config.session["downloads"]["folder"],
+            "folder_format": self.config.session["path_format"]["folder"],
+            "track_format": self.config.session["path_format"]["track"],
+            "embed_cover": self.config.session["artwork"]["embed"],
+            "embed_cover_size": self.config.session["artwork"]["size"],
+            "keep_hires_cover": self.config.session["artwork"]["keep_hires_cover"],
+            "set_playlist_to_album": self.config.session["metadata"][
+                "set_playlist_to_album"
+            ],
+            "stay_temp": self.config.session["conversion"]["enabled"],
+            "conversion": self.config.session["conversion"],
+            "concurrent_downloads": self.config.session["concurrent_downloads"],
+            "new_tracknumbers": self.config.session["metadata"][
+                "new_playlist_tracknumbers"
+            ],
+            "download_videos": self.config.session["tidal"]["download_videos"],
+            "download_booklets": self.config.session["qobuz"]["download_booklets"],
+        }
+
     def download(self):
         try:
-            arguments = {
-                "database": self.db,
-                "parent_folder": self.config.session["downloads"]["folder"],
-                "folder_format": self.config.session["path_format"]["folder"],
-                "track_format": self.config.session["path_format"]["track"],
-                "embed_cover": self.config.session["artwork"]["embed"],
-                "embed_cover_size": self.config.session["artwork"]["size"],
-                "keep_hires_cover": self.config.session["artwork"]["keep_hires_cover"],
-                "set_playlist_to_album": self.config.session["metadata"][
-                    "set_playlist_to_album"
-                ],
-                "stay_temp": self.config.session["conversion"]["enabled"],
-                "conversion": self.config.session["conversion"],
-                "concurrent_downloads": self.config.session["concurrent_downloads"],
-                "new_tracknumbers": self.config.session["metadata"][
-                    "new_playlist_tracknumbers"
-                ],
-            }
-        except KeyError as err:
+            arguments = self._get_download_args()
+        except KeyError:
+            click.secho(
+                "Updating config file... Some settings may be lost. Please run the "
+                "command again.",
+                fg="magenta",
+            )
+            self.config.update()
+            exit()
+        except Exception as err:
             click.secho(
                 "There was a problem with your config file. This happens "
                 "sometimes after updates. Run ",
@@ -210,18 +185,15 @@ class MusicDL(list):
                     click.secho(f"{item!s} is not available, skipping.", fg="red")
                     continue
 
+            item.download(**arguments)
             if isinstance(item, Track):
-                # track.download doesn't automatically tag
-                item.download(**arguments, tag=True)
-            else:
-                item.download(**arguments)
+                item.tag()
+                if arguments["conversion"]["enabled"]:
+                    item.convert(**arguments["conversion"])
 
             if self.db != [] and hasattr(item, "id"):
                 self.db.add(item.id)
 
-            # if self.config.session["conversion"]["enabled"]:
-            # item.convert(**self.config.session["conversion"])
-
     def get_client(self, source: str):
         client = self.clients[source]
         if not client.logged_in:
@@ -267,7 +239,23 @@ class MusicDL(list):
 
         :raises exceptions.ParsingError
         """
-        parsed = self.url_parse.findall(url)  # Qobuz, Tidal, Dezer
+
+        parsed = []
+
+        interpreter_urls = self.interpreter_url_parse.findall(url)
+        if interpreter_urls:
+            click.secho(
+                "Extracting IDs from Qobuz interpreter urls. Use urls "
+                "that include the artist ID for faster preprocessing.",
+                fg="yellow",
+            )
+            parsed.extend(
+                ("qobuz", "artist", extract_interpreter_url(u))
+                for u in interpreter_urls
+            )
+            url = self.interpreter_url_parse.sub("", url)
+
+        parsed.extend(self.url_parse.findall(url))  # Qobuz, Tidal, Dezer
         soundcloud_urls = self.soundcloud_url_parse.findall(url)
         soundcloud_items = [self.clients["soundcloud"].get(u) for u in soundcloud_urls]
 
@@ -284,6 +272,8 @@ class MusicDL(list):
         raise ParsingError(f"Error parsing URL: `{url}`")
 
     def handle_lastfm_urls(self, urls):
+        # https://www.last.fm/user/nathan3895/playlists/12058911
+        user_regex = re.compile(r"https://www\.last\.fm/user/([^/]+)/playlists/\d+")
         lastfm_urls = self.lastfm_url_parse.findall(urls)
         lastfm_source = self.config.session["lastfm"]["source"]
         tracks_not_found = 0
@@ -292,6 +282,11 @@ class MusicDL(list):
             global tracks_not_found
             try:
                 track = next(self.search(lastfm_source, query, media_type="track"))
+                if self.config.session["metadata"]["set_playlist_to_album"]:
+                    # so that the playlist name (actually the album) isn't
+                    # amended to include version and work tags from individual tracks
+                    track.meta.version = track.meta.work = None
+
                 playlist.append(track)
             except NoResultsFound:
                 tracks_not_found += 1
@@ -302,18 +297,20 @@ class MusicDL(list):
             title, queries = self.get_lastfm_playlist(purl)
 
             pl = Playlist(client=self.get_client(lastfm_source), name=title)
-            processes = []
+            pl.creator = user_regex.search(purl).group(1)
 
-            for title, artist in queries:
-                query = f"{title} {artist}"
-                proc = threading.Thread(
-                    target=search_query, args=(query, pl), daemon=True
-                )
-                proc.start()
-                processes.append(proc)
-
-            for proc in tqdm(processes, unit="tracks", desc="Searching"):
-                proc.join()
+            with concurrent.futures.ThreadPoolExecutor(max_workers=15) as executor:
+                futures = [
+                    executor.submit(search_query, f"{title} {artist}", pl)
+                    for title, artist in queries
+                ]
+                # only for the progress bar
+                for f in tqdm(
+                    concurrent.futures.as_completed(futures),
+                    total=len(futures),
+                    desc="Searching",
+                ):
+                    pass
 
             pl.loaded = True
             click.secho(f"{tracks_not_found} tracks not found.", fg="yellow")
@@ -368,7 +365,7 @@ class MusicDL(list):
     def preview_media(self, media):
         if isinstance(media, Album):
             fmt = (
-                "{albumartist} - {title}\n"
+                "{albumartist} - {album}\n"
                 "Released on {year}\n{tracktotal} tracks\n"
                 "{bit_depth} bit / {sampling_rate} Hz\n"
                 "Version: {version}\n"
@@ -398,7 +395,12 @@ class MusicDL(list):
         results = tuple(self.search(source, query, media_type, limit=50))
 
         def title(res):
-            return f"{res[0]+1}. {res[1].title}"
+            if isinstance(res[1], Album):
+                return f"{res[0]+1}. {res[1].album}"
+            elif isinstance(res[1], Track):
+                return f"{res[0]+1}. {res[1].meta.title}"
+            else:
+                raise NotImplementedError(type(res[1]).__name__)
 
         def from_title(s):
             num = []
@@ -422,7 +424,7 @@ class MusicDL(list):
             choice = pick(
                 tuple(enumerate(results)),
                 title=(
-                    f"{capitalize(source)} {media_type} search.\n"
+                    f"{source.capitalize()} {media_type} search.\n"
                     "Press SPACE to select, RETURN to download, ctrl-C to exit."
                 ),
                 options_map_func=title,
@@ -451,7 +453,7 @@ class MusicDL(list):
                 preview_command=from_title,
                 preview_size=0.5,
                 title=(
-                    f"{capitalize(source)} {media_type} search.\n"
+                    f"{source.capitalize()} {media_type} search.\n"
                     "SPACE - multiselection, ENTER - download, ESC - exit"
                 ),
                 cycle_cursor=True,
@@ -504,4 +506,46 @@ class MusicDL(list):
 
     def __get_source_subdir(self, source: str) -> str:
         path = self.config.session["downloads"]["folder"]
-        return os.path.join(path, capitalize(source))
+        return os.path.join(path, source.capitalize())
+
+    def prompt_creds(self, source: str):
+        """Prompt the user for credentials.
+
+        :param source:
+        :type source: str
+        """
+        if source == "qobuz":
+            click.secho(f"Enter {source.capitalize()} email:", fg="green")
+            self.config.file[source]["email"] = input()
+            click.secho(
+                f"Enter {source.capitalize()} password (will not show on screen):",
+                fg="green",
+            )
+            self.config.file[source]["password"] = md5(
+                getpass(prompt="").encode("utf-8")
+            ).hexdigest()
+
+            self.config.save()
+            click.secho(f'Credentials saved to config file at "{self.config._path}"')
+        else:
+            raise Exception
+
+    def assert_creds(self, source: str):
+        assert source in (
+            "qobuz",
+            "tidal",
+            "deezer",
+            "soundcloud",
+        ), f"Invalid source {source}"
+        if source == "deezer":
+            # no login for deezer
+            return
+
+        if source == "soundcloud":
+            return
+
+        if source == "qobuz" and (
+            self.config.file[source]["email"] is None
+            or self.config.file[source]["password"] is None
+        ):
+            self.prompt_creds(source)
