@@ -1,17 +1,22 @@
 import base64
+import contextlib
 import logging
 import os
+import re
+import sys
 from string import Formatter
 from typing import Hashable, Optional, Union
 
+import click
 import requests
 from Crypto.Cipher import AES
 from Crypto.Util import Counter
 from pathvalidate import sanitize_filename
 from requests.packages import urllib3
 from tqdm import tqdm
+from tqdm.contrib import DummyTqdmFile
 
-from .constants import LOG_DIR, TIDAL_COVER_URL
+from .constants import AGENT, LOG_DIR, TIDAL_COVER_URL
 from .exceptions import InvalidSourceError, NonStreamable
 
 urllib3.disable_warnings()
@@ -97,6 +102,20 @@ def get_quality_id(bit_depth: Optional[int], sampling_rate: Optional[int]):
         return 4
 
 
+@contextlib.contextmanager
+def std_out_err_redirect_tqdm():
+    orig_out_err = sys.stdout, sys.stderr
+    try:
+        sys.stdout, sys.stderr = map(DummyTqdmFile, orig_out_err)
+        yield orig_out_err[0]
+    # Relay exceptions
+    except Exception as exc:
+        raise exc
+    # Always restore sys.stdout/err if necessary
+    finally:
+        sys.stdout, sys.stderr = orig_out_err
+
+
 def tqdm_download(url: str, filepath: str, params: dict = None, desc: str = None):
     """Downloads a file with a progress bar.
 
@@ -118,7 +137,13 @@ def tqdm_download(url: str, filepath: str, params: dict = None, desc: str = None
 
     try:
         with open(filepath, "wb") as file, tqdm(
-            total=total, unit="iB", unit_scale=True, unit_divisor=1024, desc=desc
+            total=total,
+            unit="iB",
+            unit_scale=True,
+            unit_divisor=1024,
+            desc=desc,
+            dynamic_ncols=True,
+            # leave=False,
         ) as bar:
             for data in r.iter_content(chunk_size=1024):
                 size = file.write(data)
@@ -176,10 +201,6 @@ def init_log(path: Optional[str] = None, level: str = "DEBUG"):
     logging.basicConfig(filename=path, filemode="a", level=level)
 
 
-def capitalize(s: str) -> str:
-    return s[0].upper() + s[1:]
-
-
 def decrypt_mqa_file(in_path, out_path, encryption_key):
     # Do not change this
     master_key = "UIlTTEMmmLfGowo/UC60x2H45W6MdGgTRfo/umg4754="
@@ -232,3 +253,48 @@ def gen_threadsafe_session(
     session.mount("https://", adapter)
     session.headers.update(headers)
     return session
+
+
+def decho(message, fg=None):
+    """Debug echo the message.
+
+    :param message:
+    :param fg: ANSI color with which to display the message on the
+    screen
+    """
+    click.secho(message, fg=fg)
+    logger.debug(message)
+
+
+def extract_interpreter_url(url: str) -> str:
+    """Extract artist ID from a Qobuz interpreter url.
+
+    :param url: Urls of the form "https://www.qobuz.com/us-en/interpreter/{artist}/download-streaming-albums"
+    :type url: str
+    :rtype: str
+    """
+    session = gen_threadsafe_session({"User-Agent": AGENT})
+    r = session.get(url)
+    artist_id = re.search(r"getSimilarArtist\(\s*'(\w+)'", r.text).group(1)
+    return artist_id
+
+
+def get_container(quality: int, source: str) -> str:
+    """Get the "container" given the quality. `container` can also be the
+    the codec; both work.
+
+    :param quality: quality id
+    :type quality: int
+    :param source:
+    :type source: str
+    :rtype: str
+    """
+    if quality >= 2:
+        container = "FLAC"
+    else:
+        if source == "tidal":
+            container = "AAC"
+        else:
+            container = "MP3"
+
+    return container
